@@ -112,20 +112,23 @@ class LWLink2:
         self._callback = []
         self._group_ids = []
 
-        # Next three variables are used to synchronise responses to requests
+        # Next two variables are used to synchronise responses to requests
         self._transactions = {}
         self._response = None
 
         asyncio.ensure_future(self._consumer_handler())
 
-    async def _async_sendmessage(self, message, _retry=1):
+    async def _async_sendmessage(self, message, _retry=1, redact = False):
 
         if not self._websocket or self._websocket.closed:
             _LOGGER.info("async_sendmessage: Websocket closed, reconnecting")
             await self.async_connect()
             _LOGGER.info("async_sendmessage: Connection reopened")
 
-        _LOGGER.debug("async_sendmessage: Sending: %s", message.json())
+        if redact:
+            _LOGGER.debug("async_sendmessage: [contents hidden for security]")
+        else:
+            _LOGGER.debug("async_sendmessage: Sending: %s", message.json())
         await self._websocket.send_str(message.json())
         _LOGGER.debug("async_sendmessage: Message sent, waiting for acknowledgement from server")
         waitflag = asyncio.Event()
@@ -146,7 +149,7 @@ class LWLink2:
             return None
         else:
             _LOGGER.info("async_sendmessage: Send failed, resending message (attempt %s)", _retry + 1)
-            return await self._async_sendmessage(message, _retry + 1)
+            return await self._async_sendmessage(message, _retry + 1, redact)
 
     # Use asyncio.coroutine for compatibility with Python 3.5
     @asyncio.coroutine
@@ -176,10 +179,12 @@ class LWLink2:
                             feature_id = message["items"][0]["payload"]["featureId"]
                             feature = self.get_feature_by_featureid(feature_id)
                             value = message["items"][0]["payload"]["value"]
+                            prev_value = self.get_featureset_by_featureid(feature_id).features[feature][1]
                             self.get_featureset_by_featureid(feature_id).features[feature][1] = value
-                            _LOGGER.debug("consumer_handler: Event received (%s %s %s), calling callbacks %s", feature_id, feature, value, self._callback)
+                            cblist = [c.__name__ for c in self._callback]
+                            _LOGGER.debug("consumer_handler: Event received (%s %s %s), calling callbacks %s", feature_id, feature, value, cblist)
                             for func in self._callback:
-                                func()
+                                func(feature=feature, feature_id=feature_id, prev_value = prev_value, new_value = value)
                         else:
                             _LOGGER.warning("consumer_handler: Unhandled event message: %s", message)
                     else:
@@ -202,7 +207,7 @@ class LWLink2:
                 _LOGGER.warning("consumer_handler: unhandled exception ('{}')".format(exp))
 
     async def async_register_callback(self, callback):
-        _LOGGER.debug("async_register_callback: Register callback %s", callback)
+        _LOGGER.debug("async_register_callback: Register callback '%s'", callback.__name__)
         self._callback.append(callback)
 
     async def async_get_hierarchy(self):
@@ -372,7 +377,7 @@ class LWLink2:
         except Exception as exp:
             if (max_tries == 0) or (_retry < max_tries):
                 retry_delay = min(2 ** (_retry + 1), 120)
-                _LOGGER.warning("async_connect: Cannot connect (exception '{}'). Waiting {} seconds to retry".format(exp, retry_delay))
+                _LOGGER.warning("async_connect: Cannot connect (exception '{}'). Waiting {} seconds to retry".format(repr(exp), retry_delay))
                 await asyncio.sleep(retry_delay)
                 return await self.async_connect(max_tries, _retry + 1)
             else:
@@ -392,7 +397,7 @@ class LWLink2:
             authmess = _LWRFWebsocketMessage("user", "authenticate")
             authpayload = _LWRFWebsocketMessageItem({"token": self._authtoken, "clientDeviceId": self._device_id})
             authmess.additem(authpayload)
-            response = await self._async_sendmessage(authmess)
+            response = await self._async_sendmessage(authmess, redact = True)
             if not response["items"][0]["success"]:
                 if response["items"][0]["error"]["code"] == "200":
                     # "Channel is already authenticated" - Do nothing
@@ -425,12 +430,11 @@ class LWLink2:
         _LOGGER.debug("get_access_token_username: Requesting authentication token (using username and password)")
         authentication = {"email": self._username, "password": self._password, "version": VERSION}
         async with self._session.post(AUTH_SERVER, headers={"x-lwrf-appid": "ios-01"}, json=authentication) as req:
-            _LOGGER.debug("get_access_token_username: Received response: {}".format(await req.json()))
             if req.status == 200:
+                _LOGGER.debug("get_access_token_username: Received response: [contents hidden for security]")
+                #_LOGGER.debug("get_access_token_username: Received response: {}".format(await req.json()))
                 self._authtoken = (await req.json())["tokens"]["access_token"]
-            else:
-                _LOGGER.warning("get_access_token_username: No authentication token (status_code '{}').".format(req.status))
-                raise ConnectionError("No authentication token: {}".format(await req.text))
+
 
     # TODO check for token expiry
     async def _get_access_token_api(self):
@@ -439,14 +443,15 @@ class LWLink2:
         async with self._session.post(PUBLIC_AUTH_SERVER,
                             headers={"authorization": "basic " + self._api_token},
                             json=authentication) as req:
-            _LOGGER.debug("get_access_token_api: Received response: {}".format(await req.text))
+            _LOGGER.debug("get_access_token_api: Received response: [contents hidden for security]")
+            #_LOGGER.debug("get_access_token_api: Received response: {}".format(await req.text()))
             if req.status == 200:
-                self._authtoken = await req.json()["access_token"]
-                self._refresh_token = await req.json()["refresh_token"]
-                self._token_expiry = datetime.datetime.now() + datetime.timedelta(seconds=await req.json()["expires_in"])
+                self._authtoken = (await req.json())["access_token"]
+                self._refresh_token = (await req.json())["refresh_token"]
+                self._token_expiry = datetime.datetime.now() + datetime.timedelta(seconds=(await req.json())["expires_in"])
             else:
                 _LOGGER.warning("get_access_token_api: No authentication token (status_code '{}').".format(req.status))
-                raise ConnectionError("No authentication token: {}".format(await req.text))
+                raise ConnectionError("No authentication token: {}".format(await req.text()))
 
     #########################################################
     # Convenience methods for non-async calls
@@ -515,7 +520,8 @@ class LWLink2Public(LWLink2):
         async with self._session.get(PUBLIC_API + endpoint,
                                      headers= {"authorization": "bearer " + self._authtoken}
                                       ) as req:
-            _LOGGER.debug("async_getrequest: Received API response {} {} {}".format(req.status, await req.text(), await req.json()))
+            _LOGGER.debug("async_getrequest: Received API response {} {} {} {}".format(req.status, req.raw_headers, await req.text(), await req.json(content_type=None)))
+            #TODO if (req.status == 429) #Rate limited
             return await req.json()
 
     async def _async_postrequest(self, endpoint, body="", _retry=1):
@@ -523,7 +529,7 @@ class LWLink2Public(LWLink2):
         async with self._session.post(PUBLIC_API + endpoint,
                                       headers= {"authorization": "bearer " + self._authtoken},
                                       json=body) as req:
-            _LOGGER.debug("async_postrequest: Received API response {} {} {}".format(req.status, await req.text(), await req.json()))
+            _LOGGER.debug("async_postrequest: Received API response {} {} {} {}".format(req.status, req.raw_headers, await req.text(), await req.json(content_type=None)))
             if not(req.status == 401 and (await req.json())['message'] == 'Unauthorized'):
                 return await req.json()
         try:
@@ -533,7 +539,7 @@ class LWLink2Public(LWLink2):
                                           headers={
                                               "authorization": "bearer " + self._authtoken},
                                           json=body) as req:
-                _LOGGER.debug("async_postrequest: Received API response {} {} {}".format(req.status, await req.text(), await req.json()))
+                _LOGGER.debug("async_postrequest: Received API response {} {} {}".format(req.status, await req.text(), await req.json(content_type=None)))
                 return await req.json()
         except:
             return False
@@ -543,7 +549,7 @@ class LWLink2Public(LWLink2):
         async with self._session.delete(PUBLIC_API + endpoint,
                                      headers= {"authorization": "bearer " + self._authtoken}
                                       ) as req:
-            _LOGGER.debug("async_deleterequest: Received API response {} {} {}".format(req.status, await req.text(), await req.json()))
+            _LOGGER.debug("async_deleterequest: Received API response {} {} {} {}".format(req.status, req.raw_headers, await req.text(), await req.json(content_type=None)))
             return await req.json()
 
     async def async_get_hierarchy(self):
@@ -578,7 +584,7 @@ class LWLink2Public(LWLink2):
         payload = {"events": [{"type": "feature", "id": feature_id}],
                     "url": url,
                     "ref": ref}
-        req = await self._async_postrequest("events", payload)
+        req = await self._async_postrequest("events?New%20item=5bc4d06e87779374d29d7d9a-28-3157334318+1", payload)
         #TODO: test for req = 200
 
     async def delete_all_webhooks(self):
